@@ -31,16 +31,19 @@ struct sip_request {
 	char *met;
 	char *uri;
 	char *host;
+	char *branch;
 	struct mbuf *mb;
 	sip_send_h *sendh;
 	sip_resp_h *resph;
 	void *arg;
+	size_t sortkey;
 	enum sip_transp tp;
 	bool tp_selected;
 	bool stateful;
 	bool canceled;
 	bool provrecv;
 	uint16_t port;
+	uint16_t srcport;
 };
 
 
@@ -87,6 +90,7 @@ static void destructor(void *arg)
 	mem_deref(req->met);
 	mem_deref(req->uri);
 	mem_deref(req->host);
+	mem_deref(req->branch);
 	mem_deref(req->mb);
 }
 
@@ -144,7 +148,7 @@ static void response_handler(int err, const struct sip_msg *msg, void *arg)
 
 	req->ct = NULL;
 
-	if (!req->canceled && (err || msg->scode == 503) &&
+	if (!req->canceled && (err || (msg && msg->scode == 503)) &&
 	    (req->addrl.head || req->srvl.head)) {
 
 		err = request_next(req);
@@ -157,48 +161,91 @@ static void response_handler(int err, const struct sip_msg *msg, void *arg)
 }
 
 
-static int request(struct sip_request *req, enum sip_transp tp,
-		   const struct sa *dst)
+static int connect_handler(struct sa *src, const struct sa *dst,
+			   struct mbuf *mb, void *arg)
 {
-	struct mbuf *mb = NULL;
-	char *branch = NULL;
-	int err = ENOMEM;
-	struct sa laddr;
+	struct sip_request *req = arg;
+	struct mbuf *mbs  = NULL;
+	struct mbuf *cont = NULL;
+	int err;
 
-	req->provrecv = false;
+	if (!sa_isset(src, SA_ALL))
+		return EINVAL;
 
-	branch = mem_alloc(24, NULL);
-	mb = mbuf_alloc(1024);
+	mbuf_set_posend(mb, 0, 0);
+	mbs = mbuf_alloc(256);
+	if (!mbs)
+		return ENOMEM;
 
-	if (!branch || !mb)
-		goto out;
-
-	(void)re_snprintf(branch, 24, "z9hG4bK%016llx", rand_u64());
-
-	err = sip_transp_laddr(req->sip, &laddr, tp, dst);
+	err = req->sendh ? req->sendh(req->tp, src, dst, mbs, &cont,
+				      req->arg) : 0;
 	if (err)
 		goto out;
 
+	mbuf_set_pos(mbs, 0);
 	err  = mbuf_printf(mb, "%s %s SIP/2.0\r\n", req->met, req->uri);
 	err |= mbuf_printf(mb, "Via: SIP/2.0/%s %J;branch=%s;rport\r\n",
-			   sip_transp_name(tp), &laddr, branch);
-	err |= req->sendh ? req->sendh(tp, &laddr, dst, mb, req->arg) : 0;
+			   sip_transp_name(req->tp), src, req->branch);
+	err |= mbuf_write_mem(mb, mbuf_buf(mbs), mbuf_get_left(mbs));
 	err |= mbuf_write_mem(mb, mbuf_buf(req->mb), mbuf_get_left(req->mb));
+	if (cont) {
+		err |= mbuf_write_mem(mb, mbuf_buf(cont), mbuf_get_left(cont));
+		mem_deref(cont);
+	}
+
 	if (err)
 		goto out;
 
 	mb->pos = 0;
 
-	if (!req->stateful)
-		err = sip_send(req->sip, NULL, tp, dst, mb);
-	else
+out:
+	if (err)
+		mbuf_reset(mb);
+
+	mem_deref(mbs);
+	return err;
+}
+
+
+static int request(struct sip_request *req, enum sip_transp tp,
+		   const struct sa *dst)
+{
+	struct mbuf *mb   = NULL;
+	int err = ENOMEM;
+
+	req->provrecv = false;
+
+	mem_deref(req->branch);
+	(void)re_sdprintf(&req->branch, "z9hG4bK%016llx", rand_u64());
+	mb  = mbuf_alloc(1024);
+	if (!req->branch || !mb)
+		goto out;
+
+
+	if (!req->branch || !mb)
+		goto out;
+
+	if (req->srcport) {
+		struct sip_conncfg cfg = {.srcport = req->srcport};
+		err = sip_conncfg_set(req->sip, dst, &cfg);
+		if (err)
+			goto out;
+	}
+
+	if (!req->stateful) {
+		err = sip_send_conn(req->sip, NULL, tp, dst, mb,
+				    connect_handler, req);
+	}
+	else {
 		err = sip_ctrans_request(&req->ct, req->sip, tp, dst, req->met,
-					 branch, mb, response_handler, req);
+					 req->branch, req->host, mb,
+					 connect_handler, response_handler,
+					 req);
+	}
 	if (err)
 		goto out;
 
  out:
-	mem_deref(branch);
 	mem_deref(mb);
 
 	return err;
@@ -227,6 +274,7 @@ static int request_next(struct sip_request *req)
 		list_unlink(&rr->le);
 
 		if (req->addrl.head) {
+			dns_rrlist_sort_addr(&req->addrl, req->sortkey);
 			mem_deref(rr);
 			goto again;
 		}
@@ -259,6 +307,11 @@ static int request_next(struct sip_request *req)
 		if (req->addrl.head || req->srvl.head)
 			goto again;
 	}
+	else if (!req->stateful) {
+		req->resph = NULL;
+		terminate(req, 0, NULL);
+		mem_deref(req);
+	}
 
 	return err;
 }
@@ -281,6 +334,44 @@ static bool transp_next(struct sip *sip, enum sip_transp *tp)
 }
 
 
+static bool transp_next_srv(struct sip *sip, enum sip_transp *tp)
+{
+	enum sip_transp i;
+
+	for (i=(enum sip_transp)(*tp-1); i>SIP_TRANSP_NONE; i--) {
+		const char *srvid;
+
+		srvid = sip_transp_srvid(i);
+		if (0 == str_cmp(srvid, "???"))
+			continue;
+
+		if (!sip_transp_supported(sip, i, AF_UNSPEC))
+			continue;
+
+		*tp = i;
+		return true;
+	}
+
+	return false;
+}
+
+
+static bool transp_first(struct sip *sip, enum sip_transp *tp)
+{
+	if (!sip || !tp)
+		return false;
+
+	if (sip->tp_def != SIP_TRANSP_NONE &&
+			sip_transp_supported(sip, sip->tp_def, AF_UNSPEC)) {
+		*tp = sip->tp_def;
+		return true;
+	}
+
+	*tp = SIP_TRANSP_NONE;
+	return transp_next(sip, tp);
+}
+
+
 static bool rr_append_handler(struct dnsrr *rr, void *arg)
 {
 	struct list *lst = arg;
@@ -349,6 +440,10 @@ static bool rr_naptr_handler(struct dnsrr *rr, void *arg)
 		tp = SIP_TRANSP_TCP;
 	else if (!str_casecmp(rr->rdata.naptr.services, "SIPS+D2T"))
 		tp = SIP_TRANSP_TLS;
+	else if (!str_casecmp(rr->rdata.naptr.services, "SIP+D2W"))
+		tp = SIP_TRANSP_WS;
+	else if (!str_casecmp(rr->rdata.naptr.services, "SIPS+D2W"))
+		tp = SIP_TRANSP_WSS;
 	else
 		return false;
 
@@ -370,11 +465,17 @@ static void naptr_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 	(void)hdr;
 	(void)authl;
 
-	dns_rrlist_sort(ansl, DNS_TYPE_NAPTR);
+	dns_rrlist_sort(ansl, DNS_TYPE_NAPTR, req->sortkey);
 
 	rr = dns_rrlist_apply(ansl, NULL, DNS_TYPE_NAPTR, DNS_CLASS_IN, false,
 			      rr_naptr_handler, req);
 	if (!rr) {
+		req->tp = SIP_TRANSPC;
+		if (!transp_next_srv(req->sip, &req->tp)) {
+			err = EPROTONOSUPPORT;
+			goto fail;
+		}
+
 		err = srv_lookup(req, req->host);
 		if (err)
 			goto fail;
@@ -382,8 +483,6 @@ static void naptr_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 		return;
 	}
 
-	dns_rrlist_sort(addl, DNS_TYPE_SRV);
-
 	dns_rrlist_apply(addl, rr->rdata.naptr.replace, DNS_TYPE_SRV,
 			 DNS_CLASS_IN, true, rr_append_handler, &req->srvl);
 
@@ -397,6 +496,8 @@ static void naptr_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 		return;
 	}
 
+	dns_rrlist_sort(&req->srvl, DNS_TYPE_SRV, req->sortkey);
+
 	dns_rrlist_apply(addl, NULL, DNS_QTYPE_ANY, DNS_CLASS_IN, false,
 			 rr_cache_handler, req);
 
@@ -404,12 +505,6 @@ static void naptr_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 	if (err)
 		goto fail;
 
-	if (!req->stateful) {
-		req->resph = NULL;
-		terminate(req, 0, NULL);
-		mem_deref(req);
-	}
-
 	return;
 
  fail:
@@ -425,14 +520,12 @@ static void srv_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 	(void)hdr;
 	(void)authl;
 
-	dns_rrlist_sort(ansl, DNS_TYPE_SRV);
-
 	dns_rrlist_apply(ansl, NULL, DNS_TYPE_SRV, DNS_CLASS_IN, false,
 			 rr_append_handler, &req->srvl);
 
 	if (!req->srvl.head) {
 		if (!req->tp_selected) {
-			if (transp_next(req->sip, &req->tp)) {
+			if (transp_next_srv(req->sip, &req->tp)) {
 
 				err = srv_lookup(req, req->host);
 				if (err)
@@ -441,8 +534,7 @@ static void srv_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 				return;
 			}
 
-			req->tp = SIP_TRANSP_NONE;
-			if (!transp_next(req->sip, &req->tp)) {
+			if (!transp_first(req->sip, &req->tp)) {
 				err = EPROTONOSUPPORT;
 				goto fail;
 			}
@@ -457,6 +549,8 @@ static void srv_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 		return;
 	}
 
+	dns_rrlist_sort(&req->srvl, DNS_TYPE_SRV, req->sortkey);
+
 	dns_rrlist_apply(addl, NULL, DNS_QTYPE_ANY, DNS_CLASS_IN, false,
 			 rr_cache_handler, req);
 
@@ -464,12 +558,6 @@ static void srv_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 	if (err)
 		goto fail;
 
-	if (!req->stateful) {
-		req->resph = NULL;
-		terminate(req, 0, NULL);
-		mem_deref(req);
-	}
-
 	return;
 
  fail:
@@ -498,16 +586,12 @@ static void addr_handler(int err, const struct dnshdr *hdr, struct list *ansl,
 		goto fail;
 	}
 
+	dns_rrlist_sort_addr(&req->addrl, req->sortkey);
+
 	err = request_next(req);
 	if (err)
 		goto fail;
 
-	if (!req->stateful) {
-		req->resph = NULL;
-		terminate(req, 0, NULL);
-		mem_deref(req);
-	}
-
 	return;
 
  fail:
@@ -560,31 +644,13 @@ static int addr_lookup(struct sip_request *req, const char *name)
 }
 
 
-/**
- * Send a SIP request
- *
- * @param reqp     Pointer to allocated SIP request object
- * @param sip      SIP Stack
- * @param stateful Stateful client transaction
- * @param met      SIP Method string
- * @param metl     Length of SIP Method string
- * @param uri      Request URI
- * @param uril     Length of Request URI string
- * @param route    Next hop route URI
- * @param mb       Buffer containing SIP request
- * @param sendh    Send handler
- * @param resph    Response handler
- * @param arg      Handler argument
- *
- * @return 0 if success, otherwise errorcode
- */
-int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
-		const char *met, int metl, const char *uri, int uril,
-		const struct uri *route, struct mbuf *mb,
+static int sip_request_alloc(struct sip_request **reqp,
+		struct sip *sip, bool stateful, const char *met, int metl,
+		const char *uri, int uril, const struct uri *route,
+		enum sip_transp tp, struct mbuf *mb, size_t sortkey,
 		sip_send_h *sendh, sip_resp_h *resph, void *arg)
 {
 	struct sip_request *req;
-	struct sa dst;
 	struct pl pl;
 	int err;
 
@@ -616,21 +682,20 @@ int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
 		goto out;
 
 	req->stateful = stateful;
+	req->sortkey = sortkey;
 	req->mb    = mem_ref(mb);
 	req->sip   = sip;
 	req->sendh = sendh;
 	req->resph = resph;
 	req->arg   = arg;
 
-	if (!msg_param_decode(&route->params, "transport", &pl)) {
+	if (tp != SIP_TRANSP_NONE) {
+		req->tp = tp;
+	}
+	else if (!msg_param_decode(&route->params, "transport", &pl)) {
 
-		if (!pl_strcasecmp(&pl, "udp"))
-			req->tp = SIP_TRANSP_UDP;
-		else if (!pl_strcasecmp(&pl, "tcp"))
-			req->tp = SIP_TRANSP_TCP;
-		else if (!pl_strcasecmp(&pl, "tls"))
-			req->tp = SIP_TRANSP_TLS;
-		else {
+		req->tp = sip_transp_decode(&pl);
+		if (req->tp  == SIP_TRANSP_NONE) {
 			err = EPROTONOSUPPORT;
 			goto out;
 		}
@@ -643,8 +708,7 @@ int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
 		req->tp_selected = true;
 	}
 	else {
-		req->tp = SIP_TRANSP_NONE;
-		if (!transp_next(sip, &req->tp)) {
+		if (!transp_first(sip, &req->tp)) {
 			err = EPROTONOSUPPORT;
 			goto out;
 		}
@@ -652,6 +716,22 @@ int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
 		req->tp_selected = false;
 	}
 
+out:
+	if (err)
+		mem_deref(req);
+	else if (reqp)
+		*reqp = req;
+
+	return err;
+}
+
+
+static int sip_request_send(struct sip_request *req, struct sip *sip,
+			    const struct uri *route)
+{
+	struct sa dst;
+	int err;
+
 	if (!sa_set_str(&dst, req->host,
 			sip_transp_port(req->tp, route->port))) {
 
@@ -676,13 +756,54 @@ int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
 				 naptr_handler, req);
 	}
 
- out:
 	if (err)
 		mem_deref(req);
-	else if (reqp) {
-		req->reqp = reqp;
-		*reqp = req;
-	}
+	else if (req->reqp)
+		*req->reqp = req;
+
+	return err;
+}
+
+
+/**
+ * Send a SIP request
+ *
+ * @param reqp     Pointer to allocated SIP request object
+ * @param sip      SIP Stack
+ * @param stateful Stateful client transaction
+ * @param met      SIP Method string
+ * @param metl     Length of SIP Method string
+ * @param uri      Request URI
+ * @param uril     Length of Request URI string
+ * @param route    Next hop route URI
+ * @param mb       Buffer containing SIP request
+ * @param sortkey  Key for DNS record sorting
+ * @param sendh    Send handler
+ * @param resph    Response handler
+ * @param arg      Handler argument
+ *
+ * @return 0 if success, otherwise errorcode
+ */
+int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
+		const char *met, int metl, const char *uri, int uril,
+		const struct uri *route, struct mbuf *mb, size_t sortkey,
+		sip_send_h *sendh, sip_resp_h *resph, void *arg)
+{
+	struct sip_request *req = NULL;
+	int err;
+
+	if (!sip || !met || !uri || !route || !mb)
+		return EINVAL;
+
+	err = sip_request_alloc(&req, sip, stateful, met, metl, uri, uril,
+				route, SIP_TRANSP_NONE, mb, sortkey, sendh,
+				resph, arg);
+	if (err)
+		goto out;
+
+	req->reqp = reqp;
+	err = sip_request_send(req, sip, route);
+ out:
 
 	return err;
 }
@@ -701,7 +822,7 @@ int sip_request(struct sip_request **reqp, struct sip *sip, bool stateful,
  * @param sendh    Send handler
  * @param resph    Response handler
  * @param arg      Handler argument
- * @param fmt      Formatted SIP headers
+ * @param fmt      Formatted SIP headers and body
  *
  * @return 0 if success, otherwise errorcode
  */
@@ -752,7 +873,7 @@ int sip_requestf(struct sip_request **reqp, struct sip *sip, bool stateful,
 	mb->pos = 0;
 
 	err = sip_request(reqp, sip, stateful, met, -1, uri, -1, route, mb,
-			  sendh, resph, arg);
+			  (size_t)arg, sendh, resph, arg);
 	if (err)
 		goto out;
 
@@ -776,7 +897,7 @@ int sip_requestf(struct sip_request **reqp, struct sip *sip, bool stateful,
  * @param sendh    Send handler
  * @param resph    Response handler
  * @param arg      Handler argument
- * @param fmt      Formatted SIP headers
+ * @param fmt      Formatted SIP headers and body
  *
  * @return 0 if success, otherwise errorcode
  */
@@ -785,6 +906,7 @@ int sip_drequestf(struct sip_request **reqp, struct sip *sip, bool stateful,
 		  struct sip_auth *auth, sip_send_h *sendh, sip_resp_h *resph,
 		  void *arg, const char *fmt, ...)
 {
+	struct sip_request *req;
 	struct mbuf *mb;
 	va_list ap;
 	int err;
@@ -818,12 +940,18 @@ int sip_drequestf(struct sip_request **reqp, struct sip *sip, bool stateful,
 
 	mb->pos = 0;
 
-	err = sip_request(reqp, sip, stateful, met, -1, sip_dialog_uri(dlg),
-			  -1, sip_dialog_route(dlg), mb, sendh, resph, arg);
+	err = sip_request_alloc(&req, sip, stateful, met, -1,
+				sip_dialog_uri(dlg), -1, sip_dialog_route(dlg),
+				sip_dialog_tp(dlg),
+				mb, sip_dialog_hash(dlg), sendh, resph, arg);
 	if (err)
 		goto out;
 
- out:
+	req->reqp = reqp;
+	req->srcport = sip_dialog_srcport(dlg);
+	err = sip_request_send(req, sip, sip_dialog_route(dlg));
+
+out:
 	mem_deref(mb);
 
 	return err;
